package game.block;

import game.color.Colorizer;
import game.init.Blocks;
import game.init.Config;
import game.item.CheatTab;
import game.item.Item;
import game.material.Material;
import game.properties.IProperty;
import game.properties.PropertyBool;
import game.renderer.BlockLayer;
import game.renderer.blockmodel.ModelBlock;
import game.rng.Random;
import game.world.BlockPos;
import game.world.IWorldAccess;
import game.world.State;
import game.world.World;
import game.world.WorldServer;

public class BlockGrass extends Block implements IGrowable
{
    public static final PropertyBool SNOWY = PropertyBool.create("snowy");

    public BlockGrass()
    {
        super(Material.grass);
        this.setDefaultState(this.getBaseState().withProperty(SNOWY, Boolean.valueOf(false)));
        this.setTickRandomly();
        this.setTab(CheatTab.tabNature);
    }

    /**
     * Get the actual Block state of this Block at the given position. This applies properties not visible in the
     * metadata, such as fence connections.
     */
    public State getActualState(State state, IWorldAccess worldIn, BlockPos pos)
    {
        Block block = worldIn.getState(pos.up()).getBlock();
        return state.withProperty(SNOWY, Boolean.valueOf(block == Blocks.snow || block == Blocks.snow_layer));
    }

//    public int getBlockColor()
//    {
//        return ColorizerFoliage.getGrassColor(0.5D, 1.0D);
//    }

    public int getRenderColor(State state)
    {
    	return Colorizer.getGrassColor(0.5D, 1.0D);
//        return this.getBlockColor();
    }

    public int colorMultiplier(IWorldAccess worldIn, BlockPos pos, int renderPass)
    {
        return Colorizer.getGrassColor(worldIn, pos);
    }

    public void updateTick(WorldServer worldIn, BlockPos pos, State state, Random rand)
    {
//        if (!worldIn.client)
//        {
        if (worldIn.getLightFromNeighbors(pos.up()) < 4 && worldIn.getState(pos.up()).getBlock().getLightOpacity() > 2)
        {
        	if(Config.grassDecay)
        		worldIn.setState(pos, Blocks.dirt.getState());
        }
        else if(worldIn.getTemperatureC(pos) < 50.0f)
        {
            if (Config.grassSpread && worldIn.getLightFromNeighbors(pos.up()) >= 9)
            {
                for (int i = 0; i < 4; ++i)
                {
                    BlockPos blockpos = pos.add(rand.zrange(3) - 1, rand.zrange(5) - 3, rand.zrange(3) - 1);
                    Block block = worldIn.getState(blockpos.up()).getBlock();
                    State iblockstate = worldIn.getState(blockpos);

                    if (iblockstate.getBlock() == Blocks.dirt && iblockstate.getValue(BlockDirt.VARIANT) == BlockDirt.DirtType.DIRT && worldIn.getLightFromNeighbors(blockpos.up()) >= 4 && block.getLightOpacity() <= 2)
                    {
                        worldIn.setState(blockpos, Blocks.grass.getState());
                    }
                }
            }
        }
        else {
        	if(Config.grassDry)
        		worldIn.setState(pos, worldIn.rand.chance(20) ? Blocks.dirt.getState().withProperty(BlockDirt.VARIANT, BlockDirt.DirtType.COARSE_DIRT) : 
        				Blocks.dirt.getState());
        }
//        }
    }

    /**
     * Get the Item that this Block should drop when harvested.
     */
    public Item getItemDropped(State state, Random rand, int fortune)
    {
        return Blocks.dirt.getItemDropped(Blocks.dirt.getState().withProperty(BlockDirt.VARIANT, BlockDirt.DirtType.DIRT), rand, fortune);
    }

    /**
     * Whether this IGrowable can grow
     */
    public boolean canGrow(World worldIn, BlockPos pos, State state, boolean isClient)
    {
        return true;
    }

    public boolean canUseBonemeal(World worldIn, Random rand, BlockPos pos, State state)
    {
        return true;
    }

    public void grow(WorldServer worldIn, Random rand, BlockPos pos, State state)
    {
        BlockPos blockpos = pos.up();

        for (int i = 0; i < 128; ++i)
        {
            BlockPos blockpos1 = blockpos;
            int j = 0;

            while (true)
            {
                if (j >= i / 16)
                {
                    if (worldIn.getState(blockpos1).getBlock().material == Material.air)
                    {
                        if (rand.chance(8))
                        {
                            BlockFlower.EnumFlowerType blockflower$enumflowertype = worldIn.getBiomeGenForCoords(blockpos1).pickRandomFlower(rand, blockpos1);
                            BlockFlower blockflower = blockflower$enumflowertype.getBlockType().getBlock();
                            State iblockstate = blockflower.getState().withProperty(blockflower.getTypeProperty(), blockflower$enumflowertype);

                            if (blockflower.canBlockStay(worldIn, blockpos1, iblockstate))
                            {
                                worldIn.setState(blockpos1, iblockstate, 3);
                            }
                        }
                        else
                        {
                            State iblockstate1 = Blocks.tallgrass.getState().withProperty(BlockTallGrass.TYPE, BlockTallGrass.EnumType.GRASS);

                            if (Blocks.tallgrass.canBlockStay(worldIn, blockpos1, iblockstate1))
                            {
                                worldIn.setState(blockpos1, iblockstate1, 3);
                            }
                        }
                    }

                    break;
                }

                blockpos1 = blockpos1.add(rand.zrange(3) - 1, (rand.zrange(3) - 1) * rand.zrange(3) / 2, rand.zrange(3) - 1);

                if (worldIn.getState(blockpos1.down()).getBlock() != Blocks.grass || worldIn.getState(blockpos1).getBlock().isNormalCube())
                {
                    break;
                }

                ++j;
            }
        }
    }

    public BlockLayer getBlockLayer()
    {
        return BlockLayer.CUTOUT_MIPPED;
    }

    /**
     * Convert the BlockState into the correct metadata value
     */
    public int getMetaFromState(State state)
    {
        return 0;
    }

    protected IProperty[] getProperties()
    {
        return new IProperty[] {SNOWY};
    }

    public ModelBlock getModel(String name, State state) {
    	if(state.getValue(SNOWY))
    		return new ModelBlock("grass_side_snowed").add().nswe().d("dirt").u("grass_top");
    	else
    		return new ModelBlock("dirt").add().d().u("grass_top").tint().nswe("grass_side")
    				.add().nswe("grass_side_overlay").tint();
    }
}
