package game.gui.element;

import org.lwjgl.opengl.GL11;

import game.gui.Font;
import game.renderer.Drawing;
import game.renderer.Drawing.Offset;
import game.renderer.Drawing.Vec2i;
import game.util.CharValidator;
import game.util.ExtMath;
import game.util.Timing;
import game.util.Util;
import game.window.Button;
import game.window.Keysym;
import game.window.Window;

public class Textbox extends Element {
	public static enum Action {
		FOCUS, UNFOCUS, PREVIOUS, NEXT, FUNCTION, SEND, LINE, FORWARD, BACKWARD;
	}
	
	public static interface Callback {
		void use(Textbox elem, Action value);
	}
	
	private final Callback func;
	private final CharValidator validator;
	private final int capacity;
	private final boolean xbreak;
	private final boolean editable;
	
	private long tmr_scroll;
	private long tmr_leftmb;
	private int scrollx;
	private int scrolly;
	private int sel_start = -1;
	private int sel_end = -1;
	private int sel_drag = -1;
	private int cursorX = 0;
	private int cursorY = 0;

	private Textbox(int x, int y, int w, int h, int cap, boolean line, boolean editable, Callback callback, CharValidator validator, String text) {
		super(x, y, w, h, null);
		this.func = callback;
		this.validator = validator;
		this.capacity = cap;
		this.xbreak = !line;
		this.editable = editable;
		if(line)
			this.text_y = (this.size_y - (this.margin_y1 + this.margin_y2 + Font.YGLYPH)) / 2;
		this.setText(text);
	}

	public Textbox(int x, int y, int w, int h, int cap, boolean line, Callback callback, String text) {
		this(x, y, w, h, cap, line, true, callback, null, text);
	}

	public Textbox(int x, int y, int w, int h, int cap, boolean line, Callback callback, CharValidator validator, String text) {
		this(x, y, w, h, cap, line, true, callback, validator, text);
	}

	public Textbox(int x, int y, int w, int h, int cap, Callback callback, String text) {
		this(x, y, w, h, cap, false, true, callback, null, text);
	}

	public Textbox(int x, int y, int w, int h, String text) {
		this(x, y, w, h, Integer.MAX_VALUE, false, false, null, null, text);
	}
	
//	public void setEditable(boolean editable) {
//		this.editable = editable;
//	}
	
	protected boolean isTextCenteredX() {
		return false;
	}
	
	protected boolean isTextCenteredY() {
		return false;
	}
	
	protected boolean hasLinebreak() {
		return this.xbreak;
	}
	
	public boolean canHover() {
		return false;
	}
	
	public void setText(String str) {
		if(this.validator != null)
			str = this.validator.filter(str);
		this.text = str.length() > this.capacity ? str.substring(0, this.capacity) : str;
		this.updateText();
		this.sel_start = this.sel_end = this.sel_drag = this.text.length();
		gui_text_update_cur(this.sel_start, true);
	}
	
	public void scroll(int scr_x, int scr_y, int x, int y, boolean ctrl, boolean shift) {
		if(scr_y != 0 && this.xbreak) {
			int limit = Font.YGLYPH + this.tsize_y - (this.size_y - (this.margin_y1 + this.margin_y2));
			limit = ExtMath.clampi(limit, 0, 0x7fffffff);
			int prev = this.text_y;
			this.text_y += (scr_y < 0 ? -1 : 1) * (ctrl ? 1 : Font.YGLYPH) * this.gm.scrollLines * (shift ? 10 : 1);
			this.text_y = ExtMath.clampi(this.text_y, -limit, 0);
			if(this.sel_start >= 0)
				this.cursorY += (this.text_y - prev);
			this.r_dirty = true;
		}
		else if(scr_y != 0 || scr_x != 0) {
			int limit = Font.XGLYPH + this.tsize_x - (this.size_x - (this.margin_x1 + this.margin_x2));
			limit = ExtMath.clampi(limit, 0, 0x7fffffff);
			int prev = this.text_x;
			this.text_x += ((scr_y != 0 ? scr_y : (-scr_x)) < 0 ? -1 : 1) * (ctrl ? 1 : Font.XGLYPH) * this.gm.scrollLines * (shift ? 10 : 1);
			this.text_x = ExtMath.clampi(this.text_x, -limit, 0);
			if(this.sel_start >= 0)
				this.cursorX += (this.text_x - prev);
			this.r_dirty = true;
		}
	}
	
	public void mouse(Button btn, int x, int y, boolean ctrl, boolean shift) {
		if(btn == Button.MOUSE_LEFT) {
			if(!shift && ((Timing.tmr_current - this.tmr_leftmb) <= (((long)this.gm.dclickDelay) * 1000L))) {
				this.sel_start = this.sel_drag = 0;
				this.sel_end = this.text.length();
				this.r_dirty = true;
			}
			else {
				gui_text_select(x, y, shift);
			}
			this.tmr_leftmb = Timing.tmr_current;
		}
		else if((btn == Button.MOUSE_MIDDLE) && this.func != null) {
			this.func.use(this, Action.FUNCTION);
		}
	}
	
	public void mouserel() {
		this.scrollx = this.scrolly = 0;
	}
	
	public void drag(int x, int y) {
		gui_text_select(x, y, true);
	}
	
	public void character(char code) {
		if(this.editable) {
//			int pos = 0;
//			char chr[8];
//			utf_rwriten(chr, &pos, 8, code);
//			chr[pos] = 0;
			insertText(Character.toString(code));
		}
	}
	
	public void key(Keysym key, boolean ctrl, boolean shift) {
		if(ctrl && key == Keysym.A) {
			this.sel_start = this.sel_drag = 0;
			this.sel_end = this.text.length();
			this.r_dirty = true;
		}
		else if(ctrl && (key == Keysym.C) || (this.editable && (key == Keysym.X))) {
			if(this.sel_start >= 0 && this.sel_start != this.sel_end) { // fix empty
				// char end = this.text[this.sel_end];
				// this.text[this.sel_end] = 0;
				String str = Util.strip(this.text, this.sel_start, this.sel_end - this.sel_start, '\n', (char)0, '?');
				Window.setClipboard(str);
				// this.text[this.sel_end] = end;
				if(key == Keysym.X)
					insertText("");
			}
		}
		else if(this.editable && ctrl && key == Keysym.V) {
			insertText(Window.getClipboard());
		}
		else if(this.editable && !ctrl && key == Keysym.RETURN) {
			if(this.xbreak) {
				insertText("\n");
			}
			else if(this.func != null) {
				this.func.use(this, shift ? Action.LINE : Action.SEND);
			}
		}
		else if(this.editable && (!ctrl) && (key == Keysym.BACKSPACE || key == Keysym.DELETE)) {
			if(this.sel_start != this.sel_end) {
				insertText("");
			}
			else if(key == Keysym.DELETE && this.sel_start >= 0) {
				if(this.sel_end < this.text.length()) {
					this.sel_end += 1;
					insertText("");
				}
				else {
					this.sel_end = this.sel_start;
				}
			}
			else if(key == Keysym.BACKSPACE && this.sel_start > 0) {
				this.sel_start -= 1;
				insertText("");
			}
		}
		else if(!ctrl && (key == Keysym.RIGHT || key == Keysym.LEFT)) {
			if(key == Keysym.RIGHT && this.sel_start != this.sel_end) {
				this.sel_start = this.sel_drag = this.sel_end;
			}
			else if(key == Keysym.LEFT && this.sel_start != this.sel_end) {
				this.sel_end = this.sel_drag = this.sel_start;
			}
			if(key == Keysym.RIGHT && this.sel_start >= 0) {
				if(this.sel_end < this.text.length()) {
					this.sel_start = this.sel_drag = this.sel_end += 1;
				}
				else {
					this.sel_end = this.sel_drag = this.sel_start;
				}
				gui_text_update_cur(this.sel_end, true);
			}
			else if(key == Keysym.LEFT && this.sel_start >= 0) {
				if(this.sel_start > 0)
					this.sel_start -= 1;
				this.sel_end = this.sel_drag = this.sel_start;
				gui_text_update_cur(this.sel_end, true);
			}
		}
		else if(!ctrl && (key == Keysym.DOWN || key == Keysym.UP)) {
			if(this.xbreak) {
				if(key == Keysym.DOWN && this.sel_start != this.sel_end) {
					this.sel_start = this.sel_drag = this.sel_end;
				}
				else if(key == Keysym.UP && this.sel_start != this.sel_end) {
					this.sel_end = this.sel_drag = this.sel_start;
				}
				if(key == Keysym.DOWN && this.sel_start >= 0) {
					if(this.sel_end < this.text.length()) {
						// char ch = this.text.charAt(this.sel_end);
						// this.sel_end += 1;
						int nl = this.text.indexOf('\n', this.sel_end);
						// while(ch && ch != '\n') {
						// 	ch = utf_readn(this.text, &this.sel_end);
						// }
						this.sel_end = nl >= 0 ? nl + 1 : this.text.length();
						// else
						// 	this.sel_end -= 1;
						this.sel_start = this.sel_drag = this.sel_end;
					}
					else {
						this.sel_end = this.sel_drag = this.sel_start;
					}
					gui_text_update_cur(this.sel_end, true);
				}
				else if(key == Keysym.UP && this.sel_start >= 0) {
					// uint ch;
					if(this.sel_start > 0) {
						int nl = this.text.lastIndexOf('\n', this.sel_start);
						this.sel_start = nl >= 0 ? nl : 0;
						// do {
						// 	ch = utf_rreadn(this.text, &this.sel_start);
						// }
						// while(ch && ch != '\n');
					}
					this.sel_end = this.sel_drag = this.sel_start;
					gui_text_update_cur(this.sel_end, true);
				}
			}
			else if(this.func != null) {
				this.func.use(this, (key == Keysym.DOWN) ? Action.NEXT : Action.PREVIOUS);
			}
		}
		else if((!ctrl) && key == Keysym.TAB) {
			if(this.func != null) {
				this.func.use(this, shift ? Action.BACKWARD : Action.FORWARD);
			}
		}
	}
	
	public void select() {
		if(this.func != null) {
			this.func.use(this, Action.FOCUS);
		}
	}
	
	public void deselect() {
		this.sel_start = this.sel_end = this.sel_drag = -1;
		this.tmr_leftmb = 0L;
		this.r_dirty = true;
		if(this.func != null) {
			this.func.use(this, Action.UNFOCUS);
		}
	}
	
	public void update() {
		if((this.scrollx != 0 && !(this.xbreak)) || (this.scrolly != 0 && this.xbreak)) {
			int n;
			if(!this.xbreak)
				this.text_x += (n = (int)((float)((float)this.tmr_scroll) / 1000000.0f * 4.0f * ((float)this.scrollx)));
			else
				this.text_y += (n = (int)((float)((float)this.tmr_scroll) / 1000000.0f * 4.0f * ((float)this.scrolly)));
			if(n != 0) {
				gui_text_clamp_scroll();
				this.r_dirty = true;
			}
			if((((long)n) * 1000000L) <= this.tmr_scroll)
				this.tmr_scroll -= ((long)n) * 1000000L;
			else
				this.tmr_scroll = 0L;
			this.tmr_scroll += Timing.tmr_delta;
		}
	}
	
	public void shift(int shift_x, int shift_y) {
		super.shift(shift_x, shift_y);
		this.cursorX += shift_x;
		this.cursorY += shift_y;
	}
	
	private void gui_text_clamp_scroll() {
		int limit;
		if(this.xbreak) {
			limit = Font.YGLYPH + this.tsize_y - (this.size_y - (this.margin_y1 + this.margin_y2));
			limit = ExtMath.clampi(limit, 0, 0x7fffffff);
			this.text_y = ExtMath.clampi(this.text_y, -limit, 0);
		}
		else {
			limit = Font.XGLYPH + this.tsize_x - (this.size_x - (this.margin_x1 + this.margin_x2));
			limit = ExtMath.clampi(limit, 0, 0x7fffffff);
			this.text_x = ExtMath.clampi(this.text_x, -limit, 0);
		}
	}
	
	private void gui_text_update_cur(int offset, boolean shift) {
		int x1 = this.pos_x + this.margin_x1;
		int y1 = this.pos_y + this.margin_y1;
		int x2 = this.size_x - (this.margin_x1 + this.margin_x2);
		int y2 = this.size_y - (this.margin_y1 + this.margin_y2);
		gui_text_clamp_scroll();
		Vec2i coord = Drawing.txt_coord(offset, x1 + (this.xbreak ? 0 : this.text_x), y1 + this.text_y, 
			x1 + this.text_x, y1 + this.text_y, this.xbreak ? (this.pos_x + x2) : 0x7fffffff, 0x7fffffff, this.text);
		this.cursorX = coord.xpos;
		this.cursorY = coord.ypos;
		if(shift) {
			if(this.xbreak && this.cursorY < y1)
				this.text_y += y1 - this.cursorY;
			else if(this.xbreak && (this.cursorY + Font.YGLYPH) >= (y1 + y2))
				this.text_y -= (this.cursorY + Font.YGLYPH) - (y1 + y2);
			if(!(this.xbreak) && this.cursorX < x1)
				this.text_x += x1 - this.cursorX;
			else if(!(this.xbreak) && (this.cursorX + Font.XGLYPH) >= (x1 + x2))
				this.text_x -= (this.cursorX + Font.XGLYPH) - (x1 + x2);
			gui_text_update_cur(offset, false);
		}
		else {
			this.r_dirty = true;
		}
	}
	
	public void insertText(String str) {
		if(str == null || (this.sel_start == -1))
			return;
		str = Util.strip(str, 0, str.length(), this.xbreak ? '\n' : ' ', ' ', (char)0);
		if(this.validator != null)
			str = this.validator.filter(str);
		// plen = strlen(&sys.work_buf[1 + olen - this.sel_end]);
		// logd("t", "%d %d %d", olen, slen, plen);
		if((str.length() + this.text.length() - (this.sel_end - this.sel_start)) > this.capacity)
			return;
		this.text = this.text.substring(0, this.sel_start) + str + this.text.substring(this.sel_end);
		// memcpy(sys.work_buf, &this.text[this.sel_end], 1 + this.text.length() - this.sel_end);
		// memcpy(&this.text[this.sel_start], &sys.work_buf[1 + this.text.length() - this.sel_end], str.length());
		// memcpy(&this.text[this.sel_start + str.length()], sys.work_buf, 1 + this.text.length() - this.sel_end);
		this.sel_start += str.length();
		this.sel_end = this.sel_drag = this.sel_start;
		this.updateText();
		gui_text_update_cur(this.sel_end, true);
	}
	
	private void gui_text_select(int x, int y, boolean drag) {
		int x1 = this.pos_x + this.margin_x1;
		int y1 = this.pos_y + this.margin_y1;
		int x2 = this.size_x - (this.margin_x1 + this.margin_x2);
		int y2 = this.size_y - (this.margin_y1 + this.margin_y2);
		Offset off = Drawing.txt_offset(x, y, x1 + (this.xbreak ? 0 : this.text_x), y1 + this.text_y, 
			x1 + this.text_x, y1 + this.text_y, this.xbreak ? (this.pos_x + x2) : 0x7fffffff, 0x7fffffff, this.text);
		if(off != null) {
			this.cursorX = off.xpos;
			this.cursorY = off.ypos;
		}
		int offset = off == null ? 0 : off.offset;
		// logd("tst", "@A %d %d -. %d %d", x1, y1, x2, y2);
		// logd("tst", "@C %d %d -. %d, %d %d", x, y, offset, this.min, this.max);
		if(!drag) {
			this.sel_drag = this.sel_start = this.sel_end = offset;
		}
		else if(drag && this.sel_drag >= 0 && offset >= this.sel_drag) {
			this.sel_start = this.sel_drag;
			this.sel_end = offset;
		}
		else if(drag && this.sel_drag >= 0 && offset < this.sel_drag) {
			this.sel_start = offset;
			this.sel_end = this.sel_drag;
		}
		// logd("tst", "@S %d . %d . %d", this.sel_start, this.sel_drag, this.sel_end);
		this.r_dirty = true;
		if(x < x1)
			this.scrollx = x1 - x;
		else if(x >= (x1 + x2))
			this.scrollx = -(x - (x1 + x2));
		if(y < y1)
			this.scrolly = y1 - y;
		else if(y >= (y1 + y2))
			this.scrolly = -(y - (y1 + y2));
	}

	protected void drawBackground() {
		Drawing.drawGradient2Border(this.pos_x, this.pos_y, this.size_x, this.size_y, this.gm.style.field_top, this.gm.style.field_btm, this.gm.style.brdr_top, this.gm.style.brdr_btm);
	}
	
	protected void drawForeground(int x1, int y1, int x2, int y2) {
		Drawing.txt_draw(x1 + (this.xbreak ? 0 : this.text_x), y1 + this.text_y, 
				x1 + this.text_x, y1 + this.text_y, 
				this.xbreak ? (this.pos_x + x2) : Integer.MAX_VALUE, Integer.MAX_VALUE, this.gm.style.text_field, this.text);
		if(this.sel_start >= 0 && this.sel_end != this.sel_start)
			Drawing.txt_draw_range(this.sel_start, this.sel_end, x1 + (this.xbreak ? 0 : this.text_x), y1 + this.text_y, 
				x1 + this.text_x, y1 + this.text_y, 
				this.xbreak ? (this.pos_x + x2) : Integer.MAX_VALUE, Integer.MAX_VALUE, this.gm.style.select, this.text);
	}
	
	public void drawOverlay() {
		if(this.editable && this.sel_start >= 0 && this.sel_end == this.sel_start && this.gm.ftime() % 1.0f < 0.5f) {
			int x1 = this.pos_x + this.margin_x1;
			int y1 = this.pos_y + this.margin_y1;
			int x2 = this.size_x - (this.margin_x1 + this.margin_x2);
			int y2 = this.size_y - (this.margin_y1 + this.margin_y2);
			GL11.glScissor(x1 < 0 ? 0 : x1, (this.gm.fb_y - (y1 + y2)) < 0 ? 0 : (this.gm.fb_y - (y1 + y2)), x2 < 0 ? 0 : x2, y2 < 0 ? 0 : y2);
			GL11.glEnable(GL11.GL_SCISSOR_TEST);
			Drawing.drawRectColor(this.cursorX, this.cursorY, 1, Font.YGLYPH, this.gm.style.cursor);
			GL11.glDisable(GL11.GL_SCISSOR_TEST);
		}
	}
	
	
	
	
	
	



	public void deleteFromCursor() {
		int num = this.getNthCharFromPos() - this.sel_start;
		if(this.text.length() != 0) {
			if(this.sel_start != this.sel_end) {
				this.insertText("");
			}
			else {
//				boolean flag = num < 0;
				int i = this.sel_start + num; // flag ? this.sel_start + num : this.sel_start;
				int j = this.sel_start; // flag ? this.sel_start : this.sel_start + num;
				String s = "";

				if(i >= 0) {
					s = this.text.substring(0, i);
				}

				if(j < this.text.length()) {
					s = s + this.text.substring(j);
				}

				this.setText(s);
//				this.text = s;
//
//				if(flag) {
//					this.moveCursorBy(num);
//				}
			}
		}
	}

	public int getNthCharFromPos() {
		int i = this.sel_start;
		while(i > 0 && this.text.charAt(i - 1) != ' ') {
			--i;
		}
		return i;
	}
	
	public int getCursorPosition() {
		return this.sel_start == this.sel_end ? this.sel_start : -1;
	}
}
