package game.model;

import java.util.ArrayList;
import java.util.List;

import com.google.common.collect.Lists;

import game.renderer.blockmodel.BakedQuad;
import game.renderer.blockmodel.BreakingFour;
import game.renderer.blockmodel.ModelBlock;
import game.renderer.blockmodel.Transforms;
import game.renderer.texture.TextureAtlasSprite;
import game.world.Facing;

public class BakedModel implements IBakedModel
{
    protected final List<BakedQuad> generalQuads;
    protected final List<List<BakedQuad>> faceQuads;
    protected final boolean ambientOcclusion;
    protected final boolean gui3d;
    protected final TextureAtlasSprite texture;
    protected final Transforms cameraTransforms;

    public BakedModel(List<BakedQuad> generalQuadsIn, List<List<BakedQuad>> faceQuadsIn, boolean ambientOcclusionIn, boolean gui3dIn, TextureAtlasSprite textureIn, Transforms cameraTransformsIn)
    {
        this.generalQuads = generalQuadsIn;
        this.faceQuads = faceQuadsIn;
        this.ambientOcclusion = ambientOcclusionIn;
        this.gui3d = gui3dIn;
        this.texture = textureIn;
        this.cameraTransforms = cameraTransformsIn;
    }

    public List<BakedQuad> getFaceQuads(Facing facing)
    {
        return this.faceQuads.get(facing.ordinal());
    }

    public List<BakedQuad> getGeneralQuads()
    {
        return this.generalQuads;
    }

    public boolean isAmbientOcclusion()
    {
        return this.ambientOcclusion;
    }

    public boolean isGui3d()
    {
        return this.gui3d;
    }

    public boolean isBuiltInRenderer()
    {
        return false;
    }

    public TextureAtlasSprite getParticleTexture()
    {
        return this.texture;
    }

    public Transforms getItemCameraTransforms()
    {
        return this.cameraTransforms;
    }

    public static class Builder
    {
        private final List<BakedQuad> builderGeneralQuads;
        private final List<List<BakedQuad>> builderFaceQuads;
        private final boolean builderAmbientOcclusion;
        private TextureAtlasSprite builderTexture;
        private boolean builderGui3d;
        private Transforms builderCameraTransforms;

        public Builder(ModelBlock model)
        {
            this(model.isAmbientOcclusion(), model.isGui3d(), model.getTransform());
        }

        public Builder(IBakedModel bakedModel, TextureAtlasSprite texture)
        {
            this(bakedModel.isAmbientOcclusion(), bakedModel.isGui3d(), bakedModel.getItemCameraTransforms());
            this.builderTexture = bakedModel.getParticleTexture();

            for (Facing enumfacing : Facing.values())
            {
                this.addFaceBreakingFours(bakedModel, texture, enumfacing);
            }

            this.addGeneralBreakingFours(bakedModel, texture);
        }

        private void addFaceBreakingFours(IBakedModel bakedModel, TextureAtlasSprite texture, Facing facing)
        {
            for (BakedQuad bakedquad : bakedModel.getFaceQuads(facing))
            {
                this.addFaceQuad(facing, new BreakingFour(bakedquad, texture));
            }
        }

        private void addGeneralBreakingFours(IBakedModel p_177647_1_, TextureAtlasSprite texture)
        {
            for (BakedQuad bakedquad : p_177647_1_.getGeneralQuads())
            {
                this.addGeneralQuad(new BreakingFour(bakedquad, texture));
            }
        }

        private Builder(boolean ambientOcclusion, boolean gui3d, Transforms cameraTransforms)
        {
            this.builderGeneralQuads = Lists.<BakedQuad>newArrayList();
            this.builderFaceQuads = new ArrayList<List<BakedQuad>>(6);

            for (Facing enumfacing : Facing.values())
            {
                this.builderFaceQuads.add(Lists.<BakedQuad>newArrayList());
            }

            this.builderAmbientOcclusion = ambientOcclusion;
            this.builderGui3d = gui3d;
            this.builderCameraTransforms = cameraTransforms;
        }

        public BakedModel.Builder addFaceQuad(Facing facing, BakedQuad quad)
        {
            ((List)this.builderFaceQuads.get(facing.ordinal())).add(quad);
            return this;
        }

        public BakedModel.Builder addGeneralQuad(BakedQuad quad)
        {
            this.builderGeneralQuads.add(quad);
            return this;
        }

        public BakedModel.Builder setTexture(TextureAtlasSprite texture)
        {
            this.builderTexture = texture;
            return this;
        }

        public BakedModel makeBakedModel()
        {
            if (this.builderTexture == null)
            {
                throw new RuntimeException("Missing particle!");
            }
            else
            {
                return new BakedModel(this.builderGeneralQuads, this.builderFaceQuads, this.builderAmbientOcclusion, this.builderGui3d, this.builderTexture, this.builderCameraTransforms);
            }
        }
    }
}
